/* 
This file is part of the COPS (Cache-Oblivious Parallel SIMD Viterbi) prototype.

 COPS by Miguel M. A. Ferreira, Luis M. S. Russo and Nuno V. Roma / KDBIO & SIPS /
INESC-ID is licensed under a Creative Commons Attribution 3.0 Unported
License.  Permissions beyond the scope of this license may be
available at #email.

 The Software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 Any published media that is related with to use of the distributed
software, or derived software, must contain a reference to "Cache-Oblivious
Parallel SIMD Viterbi Decoding for Sequence Search in HMMER, Miguel Ferreira, 
Nuno Roma, Luis Russo (submitted to bmc bioinformatics)", but not in any way that
suggests that they endorse you or your use of the work.

 License available at http://creativecommons.org/licenses/by/3.0/
*/


#include <xmmintrin.h>		/* SSE	*/
#include <emmintrin.h>		/* SSE2 */

#include "hmmer.h"
#include <string.h>

typedef unsigned char byte;

#include <viterbi_cops.h>
#include <viterbi_serial.h>

/***************************************************************************/

#ifndef MAX_PARTITION
	#define MAX_PARTITION	120
#endif

#define tprintf if(0) printf
#define dprintf if(0) printf


/***************************************************************************
 *	Auxiliary functions
 ***************************************************************************/

__m128* alloc_m128_aligned64(int nvecs)
{
	void *mem;
	posix_memalign(&mem, 64, (nvecs)*sizeof(__m128));
	return (__m128*) mem;
}

void* alloc_aligned64(int nbytes)
{
	void *mem;
	posix_memalign(&mem, 64, nbytes);
	return mem;
}

int roundtop(int v, int align)
{
	if (v%align==0)	return v;
	else			return v + (align - v%align);
}

int roundfloor(int v, int align)
{
	if (v%align==0)	return v;
	else			return v - (v%align);
}


/***************************************************************************
 *	Viterbi algorithm
 ***************************************************************************/

int viterbi_cops_float_partitioned(DATA_COPS32* dcops, float* opt_res, __m128* maxlengthsptr)
{
	int L = dcops->L;
	P7_PROFILE* gm = dcops->gm;
	ESL_DSQ** ddsq = dcops->seqs;
	int M = gm->M, i, k, v, t, j;
	const int PARTITION = dcops->partition;
	float** oprmsc = dcops->rsc_msc;
	__m128* xmxEv = dcops->xmxE;
	__m128 xmxB, xmxN, xmxE, xmxC, loopN, moveN, moveC, Vinf  = _mm_set1_ps(-eslINFINITY);

	__m128  dmx[PARTITION];
	__m128  mmx[PARTITION];
	__m128  imx[PARTITION];
	__m128	xmm[12];
	float *mscore[4];
	__m128 loopC, maxlengths;

//	maskinfs= _mm_set1_ps(-eslINFINITY);	// 0xff800000, -inf, -eslINFINITY
	loopC	= _mm_set1_ps(gm->xsc[p7P_C][p7P_LOOP]);	
	// lengths -> lengths -1, para usar GT em vez de GT+Eq
	maxlengths = _mm_add_ps(*maxlengthsptr, _mm_set1_ps(1));

	moveC = _mm_set1_ps(gm->xsc[p7P_C][p7P_MOVE]);
	moveN = _mm_set1_ps(gm->xsc[p7P_N][p7P_MOVE]);
	loopN = _mm_set1_ps(gm->xsc[p7P_N][p7P_LOOP]);
	xmxC  = xmxN = Vinf;

	for (t = 0; t < M; t += PARTITION)
	{
		int t4 = t;	// do *not* divide by 4, t/4, since we are using a float* pointer
		xmxN = _mm_xor_ps(xmxN,xmxN);
		xmxB = moveN;
		for (k = 0; k < PARTITION; k++)
			dmx[k] = mmx[k] = imx[k] = Vinf;

		for (i = 1; i <= L; i++)
		{
			__m128 sc, dcv, temp, mpv, ipv, dpv;
			__m128 *ttsc = dcops->tsc_all + t*8;
			v = i-1;
			ttsc += 3;

			if (t == 0)
				xmxE = mpv = dpv = ipv = sc = dcv = Vinf;
			else {
				xmxE = xmxEv[v];
				dcv = dcops->pdcv[v];
				sc  = dcops->psc[v];
			}
			
			for (j = 0; j < 4; j++)
				mscore[j] =  oprmsc[ddsq[j][i]] + t4;
			
			for (k = 0; k < PARTITION && t+k < M; )
			{
#define EMLOAD(i)	xmm[i+8] = _mm_load_ps(mscore[i]);
				EMLOAD(0) 	EMLOAD(1) 
				EMLOAD(2) 	EMLOAD(3) 

#define MIX32(i,r)	\
	xmm[r  ] = _mm_unpacklo_ps(xmm[i], xmm[i+1]);	\
	xmm[r+1] = _mm_unpackhi_ps(xmm[i], xmm[i+1]);	
#define MIX64(i,r)	\
	xmm[r  ] = (__m128) _mm_unpacklo_pd((__m128d) xmm[i], (__m128d) xmm[i+2]);	\
	xmm[r+1] = (__m128) _mm_unpackhi_pd((__m128d) xmm[i], (__m128d) xmm[i+2]);

				MIX32(8,0)	MIX32(10,2)
				MIX64(0,4)	MIX64(1,6)

#define TRIPLETCOMPUTE(k,j)			\
{	/* Calculate new M(k), delay store */			\
	sc = _mm_max_ps(sc, _mm_add_ps(xmxB,*ttsc++));	\
	sc = _mm_add_ps(sc, xmm[j]);	\
	/* Update E */					\
	xmxE = _mm_max_ps(xmxE, sc);	\
	\
	/* Pre-emptive load of M,D,I */	\
	dpv = dmx[k];	\
	ipv = imx[k];	\
	mpv = mmx[k];	\
	\
	/* Calculate current I(k) */	\
	temp = _mm_add_ps(mpv, *ttsc++);	\
	imx[k] = _mm_max_ps(temp, _mm_add_ps(ipv, *ttsc++));	\
	\
	/* Delayed stores of M and D */	\
	mmx[k] = sc;	\
	dmx[k] = dcv;	\
	\
	/* Calculate next D, D(k+1) */	\
	sc	= _mm_add_ps(sc, *ttsc++);	\
	dcv = _mm_max_ps(sc, _mm_add_ps(dcv, *ttsc++));	\
	\
	/* Pre-emptive partial calculation of M(k+1) */	\
	sc = _mm_add_ps(mpv, *ttsc++);\
	sc = _mm_max_ps(sc, _mm_add_ps(ipv, *ttsc++));\
	sc = _mm_max_ps(sc, _mm_add_ps(dpv, *ttsc++));\
	k++;			\
}
				TRIPLETCOMPUTE(k,4)
				TRIPLETCOMPUTE(k,5)
				TRIPLETCOMPUTE(k,6)
				TRIPLETCOMPUTE(k,7)

				mscore[0] += 4; mscore[1] += 4;
				mscore[2] += 4; mscore[3] += 4;
			}

			xmxN = _mm_add_ps(xmxN, loopN);
			xmxB = _mm_add_ps(xmxN, moveN);

			if (t+k < M)
			{	v = i-1;
				xmxEv[v] = xmxE;
				dcops->pdcv[v] = dcv;
				dcops->psc [v] = sc;
			}
			else	// executed only in the last partition
			{
				__m128 lengths = _mm_set1_ps(i);
				// Greater than or Equal, GtEq. Impl as (maxL +1) > L
				__m128 cmpsgteq= _mm_cmpgt_ps(maxlengths, lengths);
				__m128 tloopC	= _mm_and_ps(loopC, cmpsgteq); 	// nullify add of loopC if after the end

				xmxC	= _mm_add_ps(xmxC, tloopC); 
				xmxC	= _mm_max_ps(xmxC, xmxE);
			}
		}
	}

	xmxC = _mm_add_ps(xmxC, moveC);
	if (opt_res != NULL)
		memmove(opt_res, &xmxC, sizeof(xmxC));

	return eslOK;
}




/***************************************************************************
 *	Setup COPS
 ***************************************************************************/

void free_cops_buffers(DATA_COPS32* dcops)
{
	int j;
	for (j = 0; j < 4; j++)
		free(dcops->seqs[j]);
	free(dcops->xmxE);
	free(dcops->pdcv);
	free(dcops->psc);
}

void alloc_cops_buffers(DATA_COPS32* dcops, int maxL)
{
	int j;
	for (j = 0; j < 4; j++)
		dcops->seqs[j] = calloc(maxL+16, sizeof(ESL_DSQ));

	dcops->psc  = alloc_m128_aligned64((maxL));
	dcops->pdcv = alloc_m128_aligned64((maxL));
	dcops->xmxE = alloc_m128_aligned64((maxL+1));
}

DATA_COPS32* p7_ViterbiCOPS_float_Create(P7_PROFILE *gm)
{
	float *tsc= gm->tsc;
	__m128 *tt;
	int M = gm->M, Mtop, k, i;
	DATA_COPS32* dr = calloc(1, sizeof(DATA_COPS32));
	Mtop = roundtop(M,4);
	dr->tsc_all = tt = alloc_m128_aligned64((Mtop+1)*8);
	dr->gm = gm;

#define SETSCORE(d,val)	*d++ = _mm_set1_ps(val);
	for (k = 0; k < M; k++, tsc += p7P_NTRANS)
	{
		SETSCORE(tt, tsc[p7P_MM]);
		SETSCORE(tt, tsc[p7P_IM]);
		SETSCORE(tt, tsc[p7P_DM]); 
		SETSCORE(tt, tsc[p7P_BM]); 
		if (k < M-1) 
		{	SETSCORE(tt, tsc[p7P_NTRANS+p7P_MI]);
			SETSCORE(tt, tsc[p7P_NTRANS+p7P_II]);
			SETSCORE(tt, tsc[p7P_NTRANS+p7P_MD]);
			SETSCORE(tt, tsc[p7P_NTRANS+p7P_DD]);
		}
		else
		{	SETSCORE(tt, -eslINFINITY);	// para MI
			SETSCORE(tt, -eslINFINITY);	// para II
			SETSCORE(tt, -eslINFINITY);	// para MD
			SETSCORE(tt, -eslINFINITY);	// para DD
		}
	}

	// fill the left-over places with neutral values
	// +4 to account for the spurious pre-emptive last loads of MM, IM, DM
	for (	; k < Mtop + 1; k++, tsc += p7P_NTRANS)
		for (i = 0; i < p7P_NTRANS; i++)
			SETSCORE(tt, -eslINFINITY);

	const int AlphaSize = gm->abc->Kp;
	dr->rsc_msc = calloc(AlphaSize+2, sizeof(int16_t*));
	float *data = (float*) alloc_aligned64(Mtop*(AlphaSize+1)*sizeof(float));

	for (k = 0; k < AlphaSize; k++)
	{
		float *rmsc = dr->rsc_msc[k] = data + k*Mtop;
		float *rsc = gm->rsc[k] + p7P_NR + p7P_MSC;	// ignore 1st values
		// #define MSC(k)   (rsc[(k)p7P_NR + p7P_MSC])
		for (i = 0; i< M; i++, rsc += p7P_NR)
			*rmsc++ = *rsc;
		// fill the left-over places with neutral values
		for (	; i < Mtop; i++)
			*rmsc++ = -eslINFINITY;
	}

	// create dummy values for k = AlphaSize, to use the padded sequences
	dr->rsc_msc[k] = data + (k)*(Mtop);
	for (i = 0; i< Mtop; i++)
		dr->rsc_msc[k][i] = -eslINFINITY;

	return dr;
}


void p7_ViterbiCOPS_float_Setup(DATA_COPS32* dcops, int maxL, int max_partition)
{
	int M = dcops->gm->M;

	if (maxL <= 10) maxL = 1000;
	if (max_partition <= 4 || max_partition % 4 != 0)
		max_partition = MAX_PARTITION;

	dcops->partition	= max_partition;
	dcops->Npartitions	= M/dcops->partition + (M % dcops->partition != 0);

	dcops->L = 0;
	dcops->allocL = maxL+16;
	alloc_cops_buffers(dcops, dcops->allocL);
}

int p7_ViterbiCOPS_float(DATA_COPS32* dcops, SEQ **seqsdb, float* results)
{
	int j, maxL = 0;
	__m128 lengths;
	float *ilengths = (float*) &lengths;

	for (j = 0; j < 4; j++)
	{	ilengths[j] = seqsdb[j]->length;
		if (ilengths[j] > maxL)
			maxL = ilengths[j];
	}

	if (maxL > dcops->allocL-10)
	{	// realloc
		dcops->allocL = ESL_MAX(dcops->allocL*2, roundtop(maxL,1024));
		free_cops_buffers(dcops);
		alloc_cops_buffers(dcops, dcops->allocL);
	}

	if (dcops->L != maxL)
	{   dcops->L  = maxL;
		p7_ReconfigLength(dcops->gm, maxL);
	}

	for (j = 0; j < 4; j++)
		// must be copied since we run N seqs in parallel up to the max length of the longest
		memcpy(dcops->seqs[j], seqsdb[j]->seq, (seqsdb[j]->length+1)*sizeof(ESL_DSQ));

	// pad sequences with k = Alphasize
	byte alphSize = dcops->gm->abc->Kp;
	for (j = 0; j < 4; j++)
		memset(dcops->seqs[j]+seqsdb[j]->length+1, alphSize, maxL-seqsdb[j]->length);

	viterbi_cops_float_partitioned(dcops, results, &lengths);
	return maxL;
}

int compare_seqs(const void *vs1, const void *vs2)
{
	SEQ *s1 = *((SEQ**) vs1);
	SEQ *s2 = *((SEQ**) vs2);
	if (s1->length == s2->length)
		return 0;
	return (s1->length > s2->length);
}




/****************************************************************
 *	Benchmark driver											*
 ****************************************************************/

/*
gcc -g -Wall -Wextra -O3 -std=gnu99 -o viterbicops-float -I. -I.. -L.. -I../../easel -L../../easel viterbi_cops_float.c viterbi_serial.c -lhmmer -leasel -lm -DCOPS_FLOAT_BENCHMARK
*/

#ifdef COPS_FLOAT_BENCHMARK

#include <string.h>
#include <sys/timeb.h>
#include "easel.h"
#include "esl_randomseq.h"
#include "esl_sqio.h"

#define STRHELPER(x) #x
#define STR(x) STRHELPER(x)

static ESL_OPTIONS options[] = {
	/* name			type		default	env	range toggles reqs incomp	help						docgroup*/
	{ "-h",	eslARG_NONE	, FALSE	, NULL, NULL , NULL, NULL, NULL, "show brief help on version and usage",0 },
	{ "-c", eslARG_NONE	, FALSE	, NULL, NULL,  NULL, NULL, NULL, "compare scores to trusted serial implementation (debug)", 0 }, 
	{ "-s",	eslARG_INT	, "42"	, NULL, NULL , NULL, NULL, NULL, "set random number seed to <n>",		0 },
	{ "-L",	eslARG_INT	, "400"	, NULL, "n>0", NULL, NULL, NULL, "length of random target seqs",		0 },
	{ "-N",	eslARG_INT	, "5000", NULL, "n>0", NULL, NULL, NULL, "number of random target seqs",		0 },
	{ "-M", eslARG_INT	, STR(MAX_PARTITION), NULL, "n>0", NULL, NULL, NULL, "maximum size for the partitions",		0 },
	{ "-R", eslARG_INT	, "1"	, NULL, "n>0", NULL, NULL, NULL, "number of test rounds",				0 },
	{	0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
};
static char usage [] = "[-options] <hmmfile>";
static char banner[] = "benchmark driver for generic Viterbi";

#define TIMEDIFF(t1,t2) ((t2.time - t1.time) + (t2.millitm - t1.millitm)*0.001)

int main(int argc, char **argv)
{
	ESL_GETOPTS	*go		= esl_getopts_CreateDefaultApp(options, 2, argc, argv, banner, usage);
	char	*hmmfile	= esl_opt_GetArg(go, 1);
	char	*seqfile	= esl_opt_GetArg(go, 2);
	ESL_STOPWATCH *w	= esl_stopwatch_Create();
	ESL_RANDOMNESS*r	= esl_randomness_Create(esl_opt_GetInteger(go, "-s"));
	ESL_ALPHABET*abc	= NULL;
	P7_HMMFILE	*hfp	= NULL;
	P7_HMM		*hmm	= NULL;
	P7_BG		*bg		= NULL;
	P7_PROFILE	*gm1, *gm2;
	int			L		= esl_opt_GetInteger(go, "-L");
	int			N		= esl_opt_GetInteger(go, "-N") / 4;
	int			MaxPart	= esl_opt_GetInteger(go, "-M");
	int			NROUNDS	= esl_opt_GetInteger(go, "-R");
	int			check	= esl_opt_GetBoolean(go, "-c");
	int			i, j;
	__m128		resdata[10];
	float 		*sc1	= (float*) resdata;
	ESL_SQFILE   *sqfp	= NULL;
	DATA_COPS32	*dcops;
	struct timeb tbstart, tbend;
	int sumlengths = 0;
	float* results = NULL;

	srand(time(NULL));
	if (p7_hmmfile_Open(hmmfile, NULL, &hfp) != eslOK) p7_Fail("Failed to open HMM file %s", hmmfile);
	if (p7_hmmfile_Read(hfp, &abc, &hmm)	 != eslOK) p7_Fail("Failed to read HMM");

	bg = p7_bg_Create(abc);
	p7_bg_SetLength(bg, L);
	gm1 = p7_profile_Create(hmm->M, abc);
	gm2 = p7_profile_Create(hmm->M, abc);
	p7_ProfileConfig(hmm, bg, gm1, L, p7_UNILOCAL);
	p7_ProfileConfig(hmm, bg, gm2, L, p7_UNILOCAL);

	dcops = p7_ViterbiCOPS_float_Create(gm1);
	p7_ViterbiCOPS_float_Setup(dcops, L+100, MaxPart); // use max L
	dcops->L = L;

	SEQ **seqsdb= calloc(4*(N+1), sizeof(SEQ*));
	int equallength = 1;

	if (esl_sqfile_OpenDigital(abc, seqfile, eslSQFILE_FASTA, NULL, &sqfp) == eslOK)
	{	// Use Sequence file
		ESL_SQ* sq = esl_sq_CreateDigital(abc);
		int dbsize = 4*N;
		int len=0;

		for (j = 0; 1; j++)
		{
			if (esl_sqio_Read(sqfp, sq) != eslOK)
				break;

			if (j > 0 && sq->n != len) 
				equallength = 0;
			len = sq->n;

			if (j > dbsize)
			{	seqsdb = realloc(seqsdb, 2*(dbsize+1)*sizeof(SEQ*));
				dbsize *= 2;
			}

			ESL_DSQ* dsq = sq->dsq;
			seqsdb[j] = malloc(sizeof(SEQ));
			seqsdb[j]->length = len;
			seqsdb[j]->seq = malloc((len+4)*sizeof(ESL_DSQ));
			memcpy(seqsdb[j]->seq, dsq, len+2);
			sumlengths += len;
			esl_sq_Reuse(sq);
		}
		N = j/4;
	}
	else
		for (i = 0; i < N; i++)
			for (j = 0; j < 4; j++)
			{
				int len = L; // - rand()%1000;
				seqsdb[i*4+j] = malloc(sizeof(SEQ));
				seqsdb[i*4+j]->seq = malloc(len+4);
				seqsdb[i*4+j]->length = len;
				esl_rsq_xfIID(r, bg->f, abc->K, len, seqsdb[i*4+j]->seq);
				sumlengths += len;
			}

	printf("ViterbiCOPS Float, model %s. ModelL.: %d, #Segms: %d, SeqL.: %d, #seqs: %d, Partition: %d, #parts: %d\n",
			hmmfile, gm1->M, (int) ceil(gm1->M/4.0), L, N*4*NROUNDS, dcops->partition, dcops->Npartitions);

	if (check) results = (float*) alloc_m128_aligned64((N+1)*2);

	ftime(&tbstart);
	if (!equallength)
	{	// Sort sequences by length
		qsort(seqsdb, N*4, sizeof(SEQ*), compare_seqs);
	}

	for (j = 0; j < NROUNDS; j++)
		for (i = 0; i < N; i++)
		{
		//	if (i % 1000 == 0) printf("Seq %d\n", i);

			p7_ViterbiCOPS_float(dcops, seqsdb+i*4, sc1);
		
			if (check) memcpy(results+i*4, sc1, 16);	// 16 bytes, 4 floats
		}
	ftime(&tbend);

	double secs = TIMEDIFF(tbstart,tbend);
	w->elapsed = w->user = secs;
	esl_stopwatch_Display(stdout, w, "# Opt CPU time: ");
	double compmillioncells = NROUNDS * (double) sumlengths * (double) hmm->M * 1e-6;
	printf("# %.0fM cells in %.1f Mc/s\n", compmillioncells, compmillioncells / secs);

	if (check)
	{	printf("Compare results against base version\n");

		for (i = 0; i < N; i++)
		{
			float sc2;
			int m1 = ESL_MAX(seqsdb[i*4+0]->length, seqsdb[i*4+1]->length);
			int m2 = ESL_MAX(seqsdb[i*4+2]->length, seqsdb[i*4+3]->length);
			int maxll = ESL_MAX(m1,m2);
			p7_ReconfigLength(gm2, maxll);

			for (j = 0; j < 4; j++)
			{
	//			p7_ReconfigLength(gm2, seqsdb[i*4+j]->length);
				p7_Viterbi_unilocal(seqsdb[i*4+j]->seq, seqsdb[i*4+j]->length, gm2, &sc2);
	
				if (results[i*4+j] != sc2)
				{	printf("Seq %d Len %4d: %f - %f\tdiff: %f\n", i*4+j, seqsdb[i*4+j]->length, 
							results[i*4+j], sc2, fabs(results[i*4+j] - sc2));
				}
			}
		}
	}
	return 0;
}

#endif

