/* 
This file is part of the COPS (Cache-Oblivious Parallel SIMD Viterbi) prototype.

 COPS by Miguel M. A. Ferreira, Luis M. S. Russo and Nuno V. Roma / KDBIO & SIPS /
INESC-ID is licensed under a Creative Commons Attribution 3.0 Unported
License.  Permissions beyond the scope of this license may be
available at #email.

 The Software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 Any published media that is related with to use of the distributed
software, or derived software, must contain a reference to "Cache-Oblivious
Parallel SIMD Viterbi Decoding for Sequence Search in HMMER, Miguel Ferreira, 
Nuno Roma, Luis Russo (submitted to bmc bioinformatics)", but not in any way that
suggests that they endorse you or your use of the work.

 License available at http://creativecommons.org/licenses/by/3.0/
*/

#include "p7_config.h"

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/timeb.h>

#include "easel.h"
#include "hmmer.h"


// Viterbi Filter SSE implementation, with Farrar scheme, 16bit integer scores. vitfilter.c
extern int p7_ViterbiFilter(const ESL_DSQ *dsq, int L, const P7_OPROFILE *om, P7_OMX *ox, float *ret_sc);

// Viterbi Filter SSE implementation, with Farrar scheme, 32bit float scores. vitscore.c
extern int p7_ViterbiScore (const ESL_DSQ *dsq, int L, const P7_OPROFILE *om, P7_OMX *ox, float *ret_sc);

#include "viterbi_serial.h"


/*****************************************************************
 *	ViterbiFilter Benchmark
 *****************************************************************/

typedef struct _dsq_cmp
{
	ESL_DSQ *seq;
	int length;
} SEQ;

/* 
gcc vitfilter_test.c viterbi_serial.c -O3 -o viterbihmmer -std=gnu99 -g -Wall -msse2  -I.. -L..  -I../impl_sse -L../impl_sse -I../../easel -L../../easel -lhmmer -lm -leasel
*/

#include "esl_stopwatch.h"
#include "esl_sqio.h"
#include "esl_randomseq.h"


static ESL_OPTIONS options[] = {
	/* name	type		default	env	range	toggles	reqs  incomp	help							docgroup*/
	{ "-h",	eslARG_NONE,FALSE, NULL, NULL,	NULL,	NULL, NULL, "show brief help",				0 },
	{ "-c",	eslARG_NONE,FALSE, NULL, NULL,	NULL,	NULL, NULL, "comp. scores to generic impl",	0 },
	{ "-s",	eslARG_INT,	"42",  NULL, NULL,	NULL,	NULL, NULL, "set random number seed to <n>",0 },
	{ "-L",	eslARG_INT,	"400", NULL, "n>0",	NULL,	NULL, NULL, "length of random target seqs",	0 },
	{ "-N",	eslARG_INT,	"5000",NULL, "n>0", NULL,	NULL, NULL, "number of random target seqs",	0 },
	{ "-R", eslARG_INT, "1"	 , NULL, "n>0",	NULL,	NULL, NULL, "number of test rounds",		0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
};
static char usage[]	= "[-options] <hmmfile>";
static char banner[] = "benchmark driver for Viterbi filter";

#define TIMEDIFF(t1,t2) ((t2.time - t1.time) + (t2.millitm - t1.millitm)*0.001)

int main(int argc, char **argv)
{
	ESL_GETOPTS	*go		= esl_getopts_CreateDefaultApp(options, 2, argc, argv, banner, usage);
	char	*hmmfile	= esl_opt_GetArg(go, 1);
	char	*seqfile	= esl_opt_GetArg(go, 2);
	ESL_STOPWATCH*w		= esl_stopwatch_Create();
	ESL_RANDOMNESS*r	= esl_randomness_Create(esl_opt_GetInteger(go, "-s"));
	ESL_ALPHABET*abc	= NULL;
	P7_HMMFILE	*hfp	= NULL;
	P7_HMM		*hmm	= NULL;
	P7_BG		*bg		= NULL;
	P7_PROFILE	*gm		= NULL;
	P7_OPROFILE	*om		= NULL;
	int			L		= esl_opt_GetInteger(go, "-L");
	int			N		= esl_opt_GetInteger(go, "-N");
	int			NROUNDS	= esl_opt_GetInteger(go, "-R");	// global
	int			check	= esl_opt_GetBoolean(go, "-c");	// global
	ESL_DSQ		*dsq	= NULL;	
	ESL_SQFILE   *sqfp	= NULL;
	int		i, j, len, newlen;
	int		sumlengths	= 0;	
	struct timeb tbstart, tbend;
	float* results = NULL;

	if (p7_hmmfile_Open(hmmfile, NULL, &hfp)!= eslOK) p7_Fail("Failed to open HMM file %s", hmmfile);
	if (p7_hmmfile_Read (hfp, &abc, &hmm)	!= eslOK) p7_Fail("Failed to read HMM");

	bg = p7_bg_Create(abc);
	p7_bg_SetLength(bg, L);
	gm = p7_profile_Create(hmm->M, abc);
	p7_ProfileConfig(hmm, bg, gm, L, p7_UNILOCAL);
	om = p7_oprofile_Create(hmm->M, abc);
	p7_oprofile_Convert(gm, om);

	SEQ **seqsdb= calloc((N+1), sizeof(SEQ*));
//	int equallength = 1;

	if (esl_sqfile_OpenDigital(abc, seqfile, eslSQFILE_FASTA, NULL, &sqfp) == eslOK)
	{	// Use Sequence file
		ESL_SQ* sq = esl_sq_CreateDigital(abc);
		int dbsize = N;
		int len=0;

		for (j = 0; 1; j++)
		{
			if (esl_sqio_Read(sqfp, sq) != eslOK)
				break;

		//	if (j > 0 && sq->n != len) equallength = 0;
			len = sq->n;

			if (j > dbsize)
			{	seqsdb = realloc(seqsdb, 2*(dbsize+1)*sizeof(SEQ*));
				dbsize *= 2;
			}

			seqsdb[j] = malloc(sizeof(SEQ));
			seqsdb[j]->length = len;
			seqsdb[j]->seq = malloc((len+2)*sizeof(ESL_DSQ));
			memcpy(seqsdb[j]->seq, sq->dsq, len+2);
			sumlengths += len;
			esl_sq_Reuse(sq);
		}
	}
	else	// Generate random sequences
		for (j = 0; j < N; j++)
		{
			int len = L; // - rand()%1000;
			seqsdb[j] = malloc(sizeof(SEQ));
			seqsdb[j]->seq = malloc(len+4);
			seqsdb[j]->length = len;
			esl_rsq_xfIID(r, bg->f, abc->K, len, seqsdb[j]->seq);
			sumlengths += len;
		}

	N = j;

	printf("ViterbiFilter HMMER, model %s. ModelL.: %d, #Segms: %d, SeqL.: %d, #seqs: %d\n",
			hmmfile, gm->M, (int) ceil(gm->M/8.0), L, N* NROUNDS);

	P7_OMX *ox	= p7_omx_Create(gm->M, 0, 0);
	if (check) results = (float*) calloc((N+1)*2, sizeof(float));

	ftime(&tbstart);

	len = 0;
	for (j = 0; j < NROUNDS; j++) 
		for (i = 0; i < N; i++)
		{
			float sc1;
			dsq = seqsdb[i]->seq;
			newlen = seqsdb[i]->length;

			if (len != newlen)	// resize if L changes
			{	len  = newlen;
				p7_oprofile_ReconfigLength(om, len);
			}

			p7_ViterbiFilter(dsq, len, om, ox, &sc1);
			if (check) results[i] = sc1;
		}

	ftime(&tbend);

	double secs = TIMEDIFF(tbstart,tbend);
	w->elapsed = w->user = secs;
	esl_stopwatch_Display(stdout, w, "# Opt CPU time: ");
	double compmillioncells = NROUNDS * (double) sumlengths * (double) hmm->M * 1e-6;
	printf("# %.0fM cells in %.1f Mc/s\n", compmillioncells, compmillioncells / secs);

	if (check)
	{	printf("Compare results against full-precision Floating-point Serial version\n");
		float sc2, sc1;
		len = 0;

		for (i = 0; i < N; i++)
		{
			dsq = seqsdb[i]->seq;
			newlen = seqsdb[i]->length;

			if (len != newlen)	// resize if L changes
			{	len  = newlen;
				p7_ReconfigLength(gm, len);
			}
			
			p7_Viterbi_unilocal(dsq, len, gm, &sc2);

			sc1 = results[i] + 1.0;	// -2.0nat optimization, Local to Unilocal mode
			if (fabs(sc1 - sc2) > 0.3)
				printf("Seq %3d-%d Len %4d: %f - %f\tdiff: %f\n", i, j, len, sc1, sc2, fabs(sc1-sc2));
		}
	}

	return 0;
}


